--***************************************************************
--* File         : WBSPIS.vhd                                   *
--* Created      : 2009/08/16 07:11:20 kawasan                  *
--* Modified     : Time-stamp: <09/10/24 17:40:11 kawasan>      *
--*                                                             *
--* WISHBONE SPI Controller                                     *
--* Slave-I/F                                                   *
--* MSB first                                                   *
--* Mode3 (http://elm-chan.org/docs/spi.html)                   *
--*                                                             *
--* Copyright 2009 Osamu Kawashima. all rights reserved         *
--***************************************************************

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity WBSPIS is
	generic (
		REG_CTRL_INIT   : std_logic_vector(7 downto 0) := x"00"
	);
	port (
		-- WISHBONE SLAVE interface:
		WB_CLK_I        : in    std_logic;
		WB_RST_I        : in    std_logic;
		WB_ADR_I        : in    std_logic_vector(7 downto 0);
		WB_DAT_I        : in    std_logic_vector(7 downto 0);
		WB_WE_I         : in    std_logic;
		WB_STB_I        : in    std_logic;
		WB_ACK_O        : out   std_logic;
		WB_DAT_O        : out   std_logic_vector(7 downto 0);
		WB_IRQ_O        : out   std_logic;

		-- Other port interface:
		SPI_SS_N        : in    std_logic;	-- Active Low
		SPI_SCLK        : in    std_logic;	-- Serial Clock
		SPI_MOSI        : in    std_logic;	-- Master-Out Slave-In
		SPI_MISO        : out   std_logic;	-- Master-In Slave-Out

		SPI_SS          : out   std_logic;
		EVENT_START     : out   std_logic;
		EVENT_STOP      : out   std_logic;
		DONE            : out   std_logic
	);
end WBSPIS;

architecture RTL of WBSPIS is

	constant ZERO32         : std_logic_vector(31 downto 0) := (others=>'0');
	constant ONE32          : std_logic_vector(31 downto 0) := (others=>'1');

	-- Address
	constant C_DATA         : std_logic_vector := x"00";
	constant C_CTRL         : std_logic_vector := x"01";
	constant C_STAT         : std_logic_vector := x"02";

	-- Bit
	constant B_DATA_EN      : integer := 0;
	constant B_DATA_IRQ_EN  : integer := 1;

	type STATE is (S_STANDBY, S_RUN, S_LATCH, S_LAST);
	signal next_state       : STATE;

	signal spi_ss_tmp       : std_logic;
	signal spi_sclk_tmp     : std_logic;
	signal spi_mosi_tmp     : std_logic;
	signal spi_ss_is        : std_logic;
	signal spi_sclk_is      : std_logic;
	signal spi_mosi_is      : std_logic;
	signal spi_ss_is2       : std_logic;
	signal spi_sclk_is2     : std_logic;
	signal spi_mosi_is2     : std_logic;

	signal reg_ctrl         : std_logic_vector(1 downto 0);

	signal dato_data        : std_logic_vector(WB_DAT_O'range);
	signal dato_ctrl        : std_logic_vector(WB_DAT_O'range);
	signal dato_stat        : std_logic_vector(WB_DAT_O'range);

	signal spi_cnt          : std_logic_vector(3 downto 0);

	signal enable           : std_logic;
	signal wr_en            : std_logic;
	signal rd_en            : std_logic;
	signal irq_enable       : std_logic;
	signal irq              : std_logic;
	signal done_is          : std_logic;

	signal spi_dat_tmp      : std_logic_vector(WB_DAT_I'range);
	signal spi_dat_next     : std_logic_vector(WB_DAT_I'range);
	signal spi_dat_rx       : std_logic_vector(WB_DAT_O'range);
	signal spi_event_start  : std_logic;
	signal spi_event_stop   : std_logic;
	signal spi_shift        : std_logic;
	signal spi_latch        : std_logic;
	signal spi_dat_latch    : std_logic;
	signal spi_miso_is      : std_logic;

begin

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if enable='0' then
				spi_ss_tmp   <= '0';
				spi_sclk_tmp <= '0';
				spi_mosi_tmp <= '0';
			else
				spi_ss_tmp   <= not SPI_SS_N;
				spi_sclk_tmp <= SPI_SCLK;
				spi_mosi_tmp <= SPI_MOSI;
			end if;
		end if;
	end process;

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if enable='0' then
				spi_ss_is   <= '0';
				spi_sclk_is <= '0';
				spi_mosi_is <= '0';
			else
				spi_ss_is   <= spi_ss_tmp;
				spi_sclk_is <= spi_sclk_tmp;
				spi_mosi_is <= spi_mosi_tmp;
			end if;
		end if;
	end process;

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if enable='0' then
				spi_ss_is2 <= '0';
				spi_sclk_is2 <= '0';
				spi_mosi_is2 <= '0';
			else
				spi_ss_is2 <= spi_ss_is;
				spi_sclk_is2 <= spi_sclk_is;
				spi_mosi_is2 <= spi_mosi_is;
			end if;
		end if;
	end process;

	spi_event_start <= '1' when spi_ss_is2='0' and spi_ss_is='1' else '0';
	spi_event_stop  <= '1' when spi_ss_is2='1' and spi_ss_is='0' else '0';

	SPI_SS <= spi_ss_is;
	EVENT_START <= spi_event_start;
	EVENT_STOP <= spi_event_stop;

	spi_shift <= '1' when spi_ss_is='1' and spi_sclk_is2='1' and spi_sclk_is='0' else '0';
	spi_latch <= '1' when spi_ss_is='1' and spi_sclk_is2='0' and spi_sclk_is='1' else '0';

	wr_en <= '1' when WB_WE_I='1' and WB_STB_I='1' else '0';
	rd_en <= '1' when WB_WE_I='0' and WB_STB_I='1' else '0';

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if enable='0' then
				spi_dat_latch <= '0';
			elsif spi_latch='1' then
				spi_dat_latch <= spi_mosi_is2;
			else
				spi_dat_latch <= spi_dat_latch;
			end if;
		end if;
	end process;

	spi_dat_next <= spi_dat_tmp(6 downto 0) & spi_dat_latch;

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if enable='0' then
				spi_dat_tmp <= (others=>'0');
				spi_miso_is <= '1';
			elsif WB_ADR_I=C_DATA and wr_en='1' then
				spi_dat_tmp <= WB_DAT_I;
				spi_miso_is <= spi_miso_is;
			elsif spi_shift='1' then
				spi_dat_tmp <= spi_dat_next;
				spi_miso_is <= spi_dat_tmp(7);
			else
				spi_dat_tmp <= spi_dat_tmp;
				spi_miso_is <= spi_miso_is;
			end if;
		end if;
	end process;

	SPI_MISO <= spi_miso_is when spi_ss_is='1' else 'Z';

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				reg_ctrl <= REG_CTRL_INIT(reg_ctrl'range);
			elsif WB_ADR_I=C_CTRL and wr_en='1' then
				reg_ctrl <= WB_DAT_I(reg_ctrl'range);
			else
				reg_ctrl <= reg_ctrl;
			end if;
		end if;
	end process;

	enable     <= '1' when reg_ctrl(B_DATA_EN)='1' and WB_RST_I='0' else '0';
	irq_enable <= reg_ctrl(B_DATA_IRQ_EN);

	-- State-Machine --
	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if enable='0' then
				next_state <= S_STANDBY;
				spi_cnt <= x"8";
				spi_dat_rx <= (others=>'0');
			else
				case next_state is
					when S_STANDBY =>
						if spi_event_start='1' then
							next_state <= S_RUN;
							spi_cnt <= x"8";
						else
							next_state <= S_STANDBY;
							spi_cnt <= spi_cnt;
						end if;
						spi_dat_rx <= spi_dat_rx;

					when S_RUN =>
						if spi_ss_is='0' then
							next_state <= S_LAST;
							spi_cnt <= spi_cnt;
							spi_dat_rx <= spi_dat_rx;
						elsif spi_cnt=0 then
							next_state <= S_LATCH;
							spi_cnt <= spi_cnt;
							spi_dat_rx <= spi_dat_next;
						elsif spi_latch='1' then
							next_state <= S_RUN;
							spi_cnt <= spi_cnt - '1';
							spi_dat_rx <= spi_dat_rx;
						end if;

					when S_LATCH =>
						next_state <= S_RUN;
						spi_cnt <= x"8";
						spi_dat_rx <= spi_dat_rx;

					when S_LAST =>
						next_state <= S_STANDBY;
						spi_cnt <= spi_cnt;
						spi_dat_rx <= spi_dat_rx;

					when others  =>
						next_state <= S_STANDBY;
						spi_cnt <= spi_cnt;
						spi_dat_rx <= spi_dat_rx;
				end case;
			end if;
		end if;
	end process;
	-- State-Machine --

	process(WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if enable='0' or (WB_ADR_I=C_DATA and rd_en='1') then
				irq <= '0';
			elsif next_state=S_LATCH then
				irq <= '1';
			else
				irq <= irq;
			end if;
		end if;
	end process;

	WB_IRQ_O <= irq and irq_enable;
	WB_ACK_O <= WB_STB_I;

	dato_data  <= spi_dat_rx;
	dato_ctrl  <= ZERO32(WB_DAT_O'left downto reg_ctrl'left+1) & reg_ctrl(reg_ctrl'range);
	dato_stat  <= ZERO32(WB_DAT_O'left downto 1) & irq;

	WB_DAT_O <= dato_data when WB_ADR_I=C_DATA else 
		    dato_ctrl when WB_ADR_I=C_CTRL else 
	            dato_stat when WB_ADR_I=C_STAT else 
		    (others=>'0');

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if enable='0' or next_state=S_LATCH then
				done_is <= '1';
			elsif WB_ADR_I=C_DATA and wr_en='1' then
				done_is <= '0';
			else
				done_is <= done_is;
			end if;
		end if;
	end process;

	DONE <= done_is;

end RTL;
